import os
import random

class MpiBench:
    'Complex mpi benchmark'
    def __init__(self, params=None):
        self.names = {'mpi'}
        self.params = {'mpi': params}

    def build(self, executor):
        'Builds the complex mpi benchmark'

        basedir = os.path.dirname(os.path.abspath(__file__))
        
        nbproc = executor.nbcores
        if nbproc < 16:
            executor.local('mpicc %s/mpi_generic.c -o /tmp/bin/mpi_generic' % basedir)
        elif nbproc <32:
            executor.local('mpicc -DSIZE=65536 %s/mpi_generic.c -o /tmp/bin/mpi_generic' % basedir)
        else:
            executor.local('mpicc -DSIZE=8192 %s/mpi_generic.c -o /tmp/bin/mpi_generic' % basedir)

        if self.params['mpi'] is None:
            self.params = {'mpi':[(16, 0, 0, 0), (0, 16, 0, 0),
                                  (0, 0, 16, 0), (0, 0, 0, 16), (4, 4, 4, 4)]}
                
        return self.params

    def run(self, bench, params, executor):
        'Runs the complex mpi benchmark'
        pcpu, pmem, pmpi, pbar = params
        delta = executor.cores('/tmp/bin/mpi_generic -c %s -m %s -n %s -b %s'
                               % (pcpu*3000, pmem, pmpi, pbar))
        return delta.strip(), 'mpigeneric-%s-%s-%s-%s' % (pcpu, pmem, pmpi, pbar)



import time
import requests
import os
import json

def get_names():
    return {'powergpu'}

class Power:
    'Monitoring GPU using Power on g5k'
    def __init__(self, lbl='pow'):
        import time
        self.label = lbl
        self.executor = None
        self.names = get_names()
        self.g5k = 'OAR_NODE_FILE' in os.environ
        self.cwd = os.path.dirname(os.path.abspath(__file__))
        now = time.localtime()
        self.df = 'nvstats-%s%s%s-%s' % (now.tm_year, str(now.tm_mon).zfill(2), str(now.tm_mday).zfill(2), self.label)
 
    def build(self, executor):
        ''
        self.executor = executor
 
    def start(self):
        'Starts the monitoring right before the benchmark'
        self.executor.local('nvidia-smi daemon -i 0 -p %s -j %s -d 1' % (self.cwd, self.label), root=True)

    def stop(self):
        'Stops the monitoring right after the benchmark'
        self.executor.local('nvidia-smi daemon -t', root=True)

    def save(self, experiment, benchname, beg_time):
        'Save the results when time is no more critical'
        filename_moj = experiment.output_file+'_mojitos'
        os.makedirs(filename_moj, exist_ok=True)
        target =  '%s/%s_%s_%s' % (filename_moj, self.executor.hostnames[0], benchname, beg_time)
        inpath = self.cwd + '/' + self.df
        outpath = inpath + '.csv'
        self.executor.local('nvidia-smi replay -i 0 -f %s -r %s' % (inpath, outpath))
        with open(outpath) as file_id:
            content = [line.split() for line in file_id.readlines()]
            beg_date = time.strftime('%Y-%m-%d', time.localtime(float(beg_time)))
            lout = []  # lines to save
            hdr = None
            for l in content:
                if hdr is None and l[0].startswith('#T'):
                    hdr = ['timestamp'] + l[1:]
                    lout.append(' '.join(hdr) + '\n')
                if l[0].startswith('#'):
                    continue
                sec = int(time.mktime(time.strptime(beg_date + ' ' + l[0] + ' CEST', '%Y-%m-%d %H:%M:%S %Z')))
                lr = [str(sec)] + l[1:]
                lout.append(' '.join(lr) + '\n')

        with open(target, 'w') as file_id:
            file_id.writelines(lout)

        # delete in the end to keep evidence in case of exception above:
        self.executor.local('rm %s %s' % (inpath, outpath))





class GpuClock:
    'nvidia-smi graphics/sm clock setting'
    def __init__(self, dummy=False, baseline=False, steps=2, zoomfrom=0, zoomto=0):
        self.dummy = dummy
        self.baseline = baseline
        self.executor = None
        self.available_frequencies = []
        self.available_frequencies_mem = []
        self.clock_mem_max = None
        self.clock_sm_min = None
        self.nsteps = steps 
        # zoomfrom/to enables full freq. granularity within a clock window:
        if zoomto != 0 and zoomfrom != zoomto:
            self.zoom = (zoomfrom, zoomto)
        else:
            self.zoom = None

    def build(self, executor):
        'Gather the available frequencies'
        self.executor = executor
        q = "nvidia-smi -i 0 --query-supported-clocks=gr --format=csv,noheader,nounits | tr '\n' ' '"
        clk_s = self.executor.local(q)
        clk = sorted([int(f) for f in clk_s.strip().split(' ')])
        self.clock_sm_min = clk[0]
        self.available_frequencies = [clk[(i*(len(clk)-1))//(self.nsteps-1)] for i in range(self.nsteps)]
        q = "nvidia-smi -i 0 --query-supported-clocks=mem --format=csv,noheader,nounits | tr '\n' ' '"
        clk_s = self.executor.local(q)
        self.available_frequencies_mem = sorted([int(f) for f in clk_s.strip().split(' ')])
        self.clock_mem_max = self.available_frequencies_mem[-1]
        if self.zoom is not None:
            # pick all available freq within the zoom window:
            clkz = [f for f in clk if self.zoom[0] <= f and f <= self.zoom[1]]
            rest = [f for f in self.available_frequencies if f < self.zoom[0] or self.zoom[1] < f]
            self.available_frequencies = sorted(clkz + rest)
        if self.dummy:
            self.available_frequencies = [self.available_frequencies[0], self.available_frequencies[-1]]
        if self.baseline:
            self.available_frequencies = [self.available_frequencies[-1]]

    def available_states(self):
        'Returns all the frequencies'
        return self.available_frequencies

    def start(self, freq):
        'Sets the right frequency range on gpu 0'
        if freq in self.available_frequencies:
            self.executor.local('nvidia-smi -i 0 -ac  %s,%s' % (self.clock_mem_max, freq), root=True)
            #self.executor.local('nvidia-smi -i 0 -lgc  %s,%s' % (self.clock_sm_min, freq), root=True)

    def stop(self, output_file=None):
        'Reverts to the maximum frequency'
        self.executor.local('nvidia-smi -i 0 -rac', root=True)
        #self.executor.local('nvidia-smi -i 0 -rgc', root=True)
            
    def get_state(self):
        'Returns the current min and max application frequencies'
        cur_min = self.clock_sm_min
        cur_max = int(self.executor.local('nvidia-smi -i 0 --query-gpu=clocks.applications.gr --format=csv,noheader,nounits'))
        return (cur_min, cur_max, self.clock_mem_max)

    def state_to_str(self):
        'Returns the current min and max frequencies as a string'
        (cur_min, cur_max, mem_max) = self.get_state()
        return '%s %s %s' % (cur_min, cur_max, mem_max)

    def get_labels(self):
        'Returns the label for frequencies'
        return ('fmin', 'fmax', 'fmemax')




import random
import os

class CpuBench:
    'Cloud-like cpu benchmark'
    def __init__(self, param_func=None, time_ref=30):
        self.names = {'cpu'}
        self.param_func = param_func
        self.time_ref = time_ref
        self.nb_processes = 1
        
    def build(self, executor):
        'Builds the cpu benchmark'

        self.nb_processes = executor.nbcores // executor.nbhosts
        if 'OAR_NODE_FILE' in os.environ:
            self.nb_processes = 2*self.nb_processes

        basedir = os.path.dirname(os.path.abspath(__file__))
            
        executor.local('gcc %s/cpu.c -o /tmp/bin/cpu' % basedir)
        ref = int(executor.local('/tmp/bin/cpu --test'))
        time_ref = 30

        if self.param_func is None:
            params = {'cpu':[(ref//3, self.time_ref),
                             ((2*ref)//3, self.time_ref),
                             (ref//2, self.time_ref),
                             (ref*2, self.time_ref)]}
        else:
            params = {'cpu': [ (res, self.time_ref) for res in self.param_func(ref)]}

        return params

    def run(self, bench, params, executor):
        'Runs the cpu benchmark'
        (load, length) = params

        delta = executor.local('/tmp/bin/cpu %s %s %s'
                               % (load, length, self.nb_processes))
        return delta.strip(), 'cpu-%s-%s' %(load, length)



import os
import random
import re
from distutils import spawn

## From https://www.mpibpc.mpg.de/16460085/bench.pdf

class GromacsBench:
    'Molecular simulation benchmark'
    def __init__(self, params=None, default_steps = 1500):
        self.names = {'gromacs'}

        self.datasets = params
        self.default_steps = default_steps
        
    def build(self, executor):
        'Install the gromacs benchmark'

        if spawn.find_executable('mdrun_mpi') is None:
            executor.hosts('apt install -y gromacs gromacs-openmpi', root=True)

        basedir = os.path.dirname(os.path.abspath(__file__))
        if self.datasets is None:
            
            executor.local('rsync %s/gromacs*.tpr /tmp/bin' % basedir)
            names = [filename[7:-4] for filename in os.listdir('/tmp/bin') if
                     filename.startswith('gromacs') and filename.endswith('.tpr')]
            self.datasets = [(name, self.default_steps) for name in names]

        else:
            for (name, _) in self.datasets:
                executor.local('rsync %s/gromacs%s.tpr /tmp/bin' % (basedir, name))

        print(self.datasets)
        return {'gromacs': self.datasets}

    def run(self, bench, params, executor):
        'Runs gromacs benchmark'
        name, niter = params
        output = executor.cores('-x GMX_MAXBACKUP=-1 mdrun_mpi -s /tmp/bin/gromacs%s.tpr -nsteps %s 2>&1' % (name, niter))
        delta = re.search('Time:[^\n]*\n', output).group(0).split()[2]
        return delta, 'gromacs-%s-%s' % (name, niter)



#!/usr/bin/python3
"""This script run HPC benchmarks with enabled monitoring

"""

import os
import random
import time
import itertools
from pathlib import Path

from functools import reduce
from execo import Process

class Executor:
    'Allow access to the platform'
    def __init__(self):
        self.mpi_host_file = '/dev/shm/mpi_host_file'
        self.mpi_core_file = '/dev/shm/mpi_core_file'
        self.mpi_options = ''
        self.hostnames = ['localhost']
        self.nbhosts = 1
        self.nbcores = os.cpu_count()
        self.sudo = 'sudo'
        self.ssh = 'ssh'
        if 'OAR_NODE_FILE' in os.environ:
            with open(os.environ['OAR_NODE_FILE']) as filename:
                content = [host.strip() for host in filename.readlines()]
                # reduce is used to keep the file order
                self.hostnames = reduce(lambda l, x: l if x in l else l+[x], content, [])
                self.nbcores = len(content)
            self.nbhosts = len(self.hostnames)
            self.mpi_options = '--mca orte_rsh_agent oarsh'
            self.sudo = 'sudo-g5k'
            self.ssh = 'oarsh'

        with open(self.mpi_host_file, 'w') as file_id:
            for host in self.hostnames:
                file_id.write(host+" slots=1\n")

        with open(self.mpi_core_file, 'w') as file_id:
            for host in self.hostnames:
                file_id.write(host+" slots=%s\n" % (self.nbcores//self.nbhosts))

    def local(self, cmd, shell=True, root=False):
        """Executes the cmd command and returns stdout after cmd exits"""
        if root:
            cmd = self.sudo+' '+cmd
        print('DEBUG:', cmd)
        proc = Process(cmd, shell=shell)
        proc.run()
        return proc.stdout

    def mpi(self, cmd, host_file, nb_processes, root=False):
        'Executes mpi cmd and returns stdout'
        if root:
            cmd = self.sudo+' '+cmd
        mpi_cmd = 'mpirun %s -np %s --machinefile %s %s' % \
                 (self.mpi_options, nb_processes, host_file, cmd)
        return self.local(mpi_cmd)

    def hosts(self, cmd, root=False):
        """Executes cmd on each host, returns rank=0 stdout"""
        return self.mpi(cmd, self.mpi_host_file, self.nbhosts, root)

    def cores(self, cmd):
        """Executes cmd on each core, returns rank=0 stdout"""
        return self.mpi(cmd, self.mpi_core_file, self.nbcores)

    def sync(self, directory):
        'Synchronize one directory accross all nodes of the experiment'
        if len(self.hostnames) > 1:
            remote = str(Path(directory).parents[0])
            for host in self.hostnames[1:]:
                self.local('oarcp -r %s %s:%s' % (directory, host, remote))

    def get_network_if(self):
        'returns the current network interface name'
        route = self.local('ip route').split()
        default_pos = route.index('default')
        dev_pos = route.index('dev', default_pos)
        return route[dev_pos+1]

################################
# Experiment
#
class Experiment:
    'Store and runs a complete experiments'
    params = {}
    def __init__(self, name, executor, benchmarks, monitors=[], leverages=[]):
        self.executor = executor
        self.output_file = '%s_%s_%s' %(name, self.executor.hostnames[0], int(time.time()))

        self.monitors = monitors
        for monitor in monitors:
            monitor.build(self.executor)

        self.hw_modes = leverages
        for leverage in leverages:
            leverage.build(self.executor)

        self.bench_suites = benchmarks
        self.build_benchs()

    def build_benchs(self):
        'Builds all the benchmarks'
        os.makedirs('/tmp/bin/', exist_ok=True)
        for bench_suite in self.bench_suites:
            tmp_params = bench_suite.build(self.executor)
            self.params.update(tmp_params)
        self.executor.sync('/tmp/bin')

    def start_monitors(self):
        'Starts all the monitors'
        for monitor in self.monitors:
            monitor.start()
    def stop_monitors(self):
        'Stops all the monitors'
        for monitor in self.monitors:
            monitor.stop()
    def save_monitors(self, benchname, beg_time):
        'Save all the monitored values'
        for monitor in self.monitors:
            monitor.save(self, benchname, beg_time)

    def get_hw_actions(self):
        'Returns all possible combination of leverages'
        possibilities = [[(leverage, state) for state in leverage.available_states()]
                         for leverage in self.hw_modes]
        return itertools.product(*possibilities)
    def start_actions(self, actions):
        'Put each leverage in the right state'
        for (leverage, state) in actions:
            leverage.start(state)

    def stop_actions(self, actions, pattern):
        'Revert to stable state'
        for (leverage, _) in actions:
            leverage.stop(pattern)

    def get_all_benchnames(self):
        'returns the list of benchmark names'
        return set().union(*[suite.names for suite in self.bench_suites])
    def get_params(self, bench):
        'returns the list of parameters for a specific benchmark'
        return self.params[bench]
    def run_bench(self, bench, param):
        'runs a benchmark with param and returns stdout'
        for bench_suite in self.bench_suites:
            if bench in bench_suite.names:
                return bench_suite.run(bench, param, self.executor)

    def do_watermark(self):
        print('Watermark 30s')
        mark = 'cat /dev/zero > /dev/null & pid=$!; sleep 5; kill -9 ${pid}'
        for _ in range(3):
            time.sleep(3.3)
            self.executor.cores(mark)
        time.sleep(5)
            
    def print_header(self):
        'Prints the initial header in the experimental output file'
        if os.path.exists(self.output_file):
            return
        res = 'hostname fullname nproc duration startTime endTime '
        for leverage in self.hw_modes:
            res += ' '.join(leverage.get_labels())
        res += ' hostlist\n'
        with open(self.output_file, 'w') as output_file:
            output_file.write(res)

    def monitor_bench(self, bench, param, watermark=False):
        'Runs and monitor the benchmark and then save the results'
        output_time = ''
        self.print_header()
        hw_state = ' '.join([leverage.state_to_str() for leverage in self.hw_modes])
        print('\nstarting ', self.output_file, bench, param, 'at', hw_state)

        while output_time == '':
            time.sleep(5)

            beg_time = int(time.time())
            self.start_monitors()
            if watermark:
                self.do_watermark()
            output_time, benchname = self.run_bench(bench, param)
            if watermark:
                self.do_watermark()
            self.stop_monitors()
            end_time = int(time.time())

        result = '%s %s %s %s %s %s ' % (self.executor.hostnames[0], benchname,
                                            self.executor.nbcores, output_time,
                                            beg_time, end_time)
        if hw_state != '':
            result += hw_state+' '
        result += ';'.join(self.executor.hostnames)
        result += '\n'
        with open(self.output_file, 'a') as output_file:
            output_file.write(result)

        self.save_monitors(benchname, beg_time)
        return "%s_%s/%s_%s_%s" % (self.output_file, "%s",
                                   self.executor.hostnames[0],
                                   benchname, beg_time)

def run_experiment(name, benchmarks, leverages=[], monitors=[], sweep=False, times=1, watermark=False):
    """Typical experiment"""
    executor = Executor()

    expe = Experiment(name, executor, benchmarks,
                      monitors=monitors, leverages=leverages)

    for _ in range(times):
        for bench in expe.get_all_benchnames():
            for param in expe.get_params(bench):
                if callable(param):
                    param = param()
                for actions in expe.get_hw_actions():
                    expe.start_actions(actions)
                    pattern = expe.monitor_bench(bench, param, watermark=watermark)
                    expe.stop_actions(actions, pattern)



import os
import re

def standard_parameters(nbproc):
    if nbproc < 16:
        return {'ep':'C', 'bt':'A', 'sp':'A', 'cg':'B',
                'is':'C', 'ft':'B', 'lu':'A', 'mg':'C'}
    if nbproc < 64:
        return  {'ep':'C', 'bt':'B', 'sp':'B', 'cg':'C',
                 'is':'D', 'ft':'C', 'lu':'B', 'mg':'D'}
    if nbproc < 256:
        return  {'ep':'D', 'bt':'C', 'sp':'C', 'cg':'D',
                 'is':'D', 'ft':'C', 'lu':'C', 'mg':'D'}
    # nbproc >= 256
    return {'ep':'D', 'bt':'D', 'sp':'D', 'cg':'D',
            'is':'E', 'ft':'E', 'lu':'D', 'mg':'D'}
    

class NpbBench:
    'Nas Parallel Benchmark'
    def __init__(self, names={'ep', 'bt', 'sp', 'cg', 'is', 'ft', 'lu', 'mg'},
                 options=None):
        self.names = names
        self.params = options

    def build(self, executor):

        basedir = os.path.dirname(os.path.abspath(__file__))
        
        'Builds NPB benchmark'
        executor.local('tar xfC %s/NPB3.4-MPI.tgz /tmp/' % basedir, shell=False)
        executor.local('cp /tmp/NPB3.4-MPI/config/make.def.template /tmp/NPB3.4-MPI/config/make.def ')
        nbproc = executor.nbcores

        if self.params is None:
            self.params = standard_parameters(nbproc)

        for bench in self.names:
            executor.local('cd /tmp/NPB3.4-MPI/; make %s CLASS=%s'
                           %(bench, self.params[bench]))

        executor.local('cp /tmp/NPB3.4-MPI/bin/* /tmp/bin/')

        for key in self.params:
            self.params[key] = [(self.params[key], nbproc)]
        return self.params

    def run(self, bench, params, executor):
        """Runner for NPB benchmarks """
        classtype, nbproc = params
        output = executor.cores('/tmp/bin/%s.%s.x' % (bench, classtype))
        execution_time = float(re.search(' Time in seconds = *(.*)', output).group(1))
        benchname = bench+'-'+classtype+'-'+str(nbproc)
        return execution_time, benchname



import os
import random
import time

class GenericBench:
    'Wide coverage benchmark, from memory to processor and network'
    def __init__(self, step_duration = 4, step_size = 25):
        self.names = {'generic'}
        self.duration = step_duration
        self.step = step_size
        self.remote = "localhost"

    def build(self, executor):
        'Builds the wide coverage benchmark'
        if len(executor.hostnames) > 1:
            self.remote = executor.hostnames[-1]

        basedir = os.path.dirname(os.path.abspath(__file__))
        for source in ["ub_cpu_cu", "ub_cpu_alu",
                       "ub_cpu_rand", "setcpulatency", "ub_mem_access"]:
            executor.local('gcc %s/genericbench/%s.c -o /tmp/bin/%s' %
                           (basedir, source, source))
        executor.local('cp %s/genericbench/uc-generic.sh /tmp/bin' % basedir)
        executor.hosts('apt install -y iperf3 cpulimit', root=True)
        executor.local('chmod go+w /dev/cpu_dma_latency', root=True)
        params = {'generic': [None]}

        return params

    def run(self, bench, params, executor):
        'Runs the wide coverage benchmark'

        starting_time = time.time()
        throughput = executor.local("/tmp/bin/uc-generic.sh %s %s" %
                                    (self.duration, self.remote))
        delta = time.time() - starting_time
        return delta, 'generic'
